import logging
import logging.config

__doc__ = """
Logger
======

Based on:
https://docs.python.org/3/howto/logging-cookbook.html#formatting-times-using-utc-gmt-via-configuration

https://realpython.com/lessons/logger-dictionary/

The Galaxy Project has the same base:
https://docs.galaxyproject.org/en/release_20.05/admin/config_logging.html

Collected logger for this repo and to abstract out the config for logger

https://docs.python.org/3/library/logging.html#logrecord-attributes
"""


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class UTCFormatter(logging.Formatter):
    """Force UTC time in log messages"""

    converter = logging.time.gmtime


class UTCColorFormatter(logging.Formatter):
    """Force UTC time in log messages and add color to log level"""

    converter = logging.time.gmtime
    # debug
    cyan = "\033[36m"
    # info
    green = "\033[32m"
    # warning
    yellow = "\033[33m"
    # error
    red = "\033[31m"
    # critical
    bold_red = "\033[91m"
    reset = "\033[0m"

    start_str = "[%(asctime)s.%(msecs)03dZ]"

    end_str = ": %(name)s : %(message)s"

    datefmt = "%Y-%m-%dT%H:%M:%S"

    FORMATS = {
        logging.DEBUG: f"{start_str} {cyan}%(levelname)s{reset} {end_str} ",
        logging.INFO: f"{start_str} {green}%(levelname)s{reset} {end_str}",
        logging.WARNING: f"{start_str} {yellow}%(levelname)s{reset} {end_str}",
        logging.ERROR: f"{start_str} {red}%(levelname)s{reset} {end_str}",
        logging.CRITICAL: f"{start_str} {bold_red}%(levelname)s{reset} {end_str}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt=self.datefmt)
        return formatter.format(record)


class HideMessageStartswith(logging.Filter):
    """Filter out the schema from the log

    from: https://stackoverflow.com/a/879937
    """

    def __init__(self, text):
        self.text = text

    def filter(self, record):
        return not record.getMessage().startswith(self.text)



from odh_core.graphql_client import GraphQLClient  # pyright: ignore[reportUnusedImport]
from odh_core.logging_settings import (
    LOGGING_AUTO,
    LOGGING_DEV,
    LOGGING_PROD,
)  # pyright: ignore[reportUnusedImport]
from odh_core.ms_odata import ms_odata  # pyright: ignore[reportUnusedImport]
from odh_core.phonenumber import parse_nr  # pyright: ignore[reportUnusedImport]
from odh_core.ssn import validate_ssn  # pyright: ignore[reportUnusedImport]

__version__ = "0.1.20"



from personnummer import personnummer
from personnummer.personnummer import PersonnummerException


def validate_ssn(ssn, country="SE", fail_as_none=False):
    """Validates SSN and normalizes it.

    Supports
        - SE, including coordination numbers

    Args:
        ssn (str): SSN to validate.
        country (str): Country code, ISO 3166-1 alpha-2.
            default: "SE"

    Returns:
        str or None: Normalized SSN.
    """
    if country == "SE":
        try:
            pn = personnummer.parse(ssn)
            if pn.valid():
                return pn.format(long_format=True)
        except PersonnummerException as e:
            if fail_as_none:
                return None
            else:
                raise ValueError(f"Invalid SSN: {ssn}") from e
    else:
        raise ValueError(f"Unsupported country: {country}")



import logging
from urllib.parse import urlsplit, urlunsplit
from datetime import datetime, timezone
from jose import jwt
import aiohttp
import asyncio

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class MicrosoftException(Exception):
    """General Microsoft Exception, Transport error, server error, etc."""

    pass


class GraphError(Exception):
    """Exception to catch Microsoft Graph API errors, user code did something wrong"""

    pass


def relative_url(url: str) -> str:
    """Convert absolute url to relative url"""
    return urlunsplit(urlsplit(url)._replace(scheme="", netloc=""))


class ms_odata:
    def __init__(
        self,
        CLIENT_ID: str,
        TENANT_ID: str,
        CLIENT_SECRET: str,
        scope: str = "https://graph.microsoft.com/.default",
        b2c=False,
        base_url="https://graph.microsoft.com/",
    ) -> None:
        self.CLIENT_ID = CLIENT_ID
        self.TENANT_ID = TENANT_ID
        self.CLIENT_SECRET = CLIENT_SECRET
        self.scope = scope
        self.b2c = b2c
        self.token = {}
        self.session = aiohttp.ClientSession(base_url=base_url, trust_env=True)

    async def get_token(self):
        """Get the token from Microsoft Graph"""

        # TENANT_ID can be guid or domain name (e.g. contoso.onmicrosoft.com)
        url = f"https://login.microsoftonline.com/{self.TENANT_ID}/oauth2/v2.0/token"

        # Set the headers
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        # Set the payload

        payload = {
            "grant_type": "client_credentials",
            "client_id": self.CLIENT_ID,
            "client_secret": self.CLIENT_SECRET,
            # msgraph only needs this scope for now
            "scope": self.scope,
        }

        # Make the request to get the token
        # TODO: remove the session
        async with aiohttp.ClientSession() as session:
            async with session.post(url, headers=headers, data=payload) as resp:
                token_json = await resp.json()
                if not resp.ok:
                    raise MicrosoftException(
                        f"Request to {url} failed with {resp.status}"
                    )
                # decode jwt token to AzureADToken
                tokendecoded = jwt.get_unverified_claims(token_json["access_token"])

                self.token["access_token"] = token_json["access_token"]
                self.token["expires_on"] = datetime.fromtimestamp(
                    tokendecoded["exp"], tz=timezone.utc
                )
                self.token["not_before"] = datetime.fromtimestamp(
                    tokendecoded["nbf"], tz=timezone.utc
                )
                self.token["resource"] = tokendecoded["aud"]

                log.info(
                    f'Claimed new token, expires on {self.token["expires_on"]},'
                    f" for {self.TENANT_ID}"
                )

    async def update_token_header(self):
        # TODO: merge this in to get_token
        await self.get_token()
        # create a session with required headers and retry policy

        # save the access token in settings object?

        self.session.headers.update(
            {
                "Authorization": f'Bearer {self.token["access_token"]}',
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        )

    async def get_one(
        self,
        resource: str,
        filter: str = None,
        # add support to have str och obj as input
        # have default select based on what resource it is
        expand: str = None,
        select: str = None,
        deltaLink: str = None,
        nextLink: str = None,
        endpoint: str = "v1.0",
    ):
        """Function to do Odata querys against Microsfts Graph API
        to use delta queries call a /delta resource. @odata.deltaLink will be
        returned as part of the response and has to be set un subsequent calls.
        If the initial query retuns nextLink, the link will be followed and complete
        data returned

        """
        request_retry = 0
        while request_retry < 5:
            url = None
            try:
                # deltalink and nextlink are absolute url and can't base_url
                # per odata spec domain can't be changed
                if deltaLink:
                    url = relative_url(deltaLink)
                elif nextLink:
                    url = relative_url(nextLink)
                else:
                    url = f"/{endpoint}/{resource}"
                    parameters = []
                    if filter:
                        parameters.append(f"$filter={filter}")
                    if select:
                        parameters.append(f"$select={select}")
                    if expand:
                        parameters.append(f"$expand={expand}")
                    if parameters:
                        url = f"{url}?{'&'.join(parameters)}"
                # make the request
                async with self.session.get(url) as resp:
                    # if we get 200, return the data
                    if resp.status == 200:
                        return await resp.json()
                    # if we get 401, refresh the token and retry
                    elif resp.status == 401:
                        # refresh the token
                        await self.update_token_header()
                        request_retry += 1
                    # if we get 404, return empty
                    elif resp.status == 404:
                        return {}
                    # if we get 429, wait and retry
                    elif resp.status == 429:
                        delay = float(resp.headers["retry-after"])
                        log.info(
                            f"Request to {url} failed with 429, retrying in {delay} seconds"
                        )
                        await asyncio.sleep(delay)
                        request_retry += 1
                    # if we get 503, wait and retry
                    elif resp.status == 503:
                        delay = 5
                        log.info(
                            f"Request to {url} failed with 503, retrying in {delay} seconds"
                        )
                        await asyncio.sleep(delay)
                        request_retry += 1
                    # if we get anything else, log it and return empty
                    else:
                        log.error(
                            f"Request to {url} failed with {resp.status} {resp.reason}"
                        )
                        return {}
            except aiohttp.ClientError as e:
                log.error(f"Request to {url} failed with {e}")
                return {}

    async def get_all(
        self,
        resource: str,
        filter: str = None,
        expand: str = None,
        select: str = None,
        endpoint: str = "v1.0",
        deltaLink: str = None,
    ):
        nextLink = None
        # get the odata
        odata_records = []
        while True:
            graphdata = await self.get_one(
                resource=resource,
                select=select,
                filter=filter,
                expand=expand,
                nextLink=nextLink,
                deltaLink=deltaLink,
                endpoint=endpoint,
            )
            # if we get an empty response, return
            if not graphdata:
                return odata_records
            # if we get a single response return it
            if graphdata.get("value") is None:
                return graphdata
            # if we get several responses, add the values to the list
            for value in graphdata["value"]:
                odata_records.append(value)
            # if we get an empty response, return
            if "@odata.nextLink" in graphdata:
                nextLink = graphdata["@odata.nextLink"]
            elif "@odata.deltaLink" in graphdata:
                deltaLink = graphdata["@odata.deltaLink"]
                return odata_records, deltaLink
            else:
                return odata_records, None

    # # not working yet
    # async def get(
    #     self,
    #     resource: str,
    #     filter: str = None,
    #     expand: str = None,
    #     select: str = None,
    #     endpoint: str = "v1.0",
    #     deltaLink: str = None,
    # ):
    #     raise NotImplementedError
    #     nextLink = None
    #     # # get the odata
    #     # odata_records = []
    #     while True:
    #         graphdata = await self.get_one(
    #             resource=resource,
    #             select=select,
    #             filter=filter,
    #             expand=expand,
    #             deltaLink=deltaLink,
    #             endpoint=endpoint,
    #         )
    #         # if we get an empty response, stop
    #         if not graphdata:
    #             raise StopAsyncIteration
    #             # return odata_records
    #         # if we get a single response return it
    #         if graphdata.get("value") is None:
    #             yield graphdata
    #         # if we get several responses, add the values to the list
    #         for value in graphdata["value"]:
    #             yield value
    #             # odata_records.append(value)

    #         # if we get a nextLink, follow it
    #         if "@odata.nextLink" in graphdata:
    #             nextLink = graphdata["@odata.nextLink"]

    #         elif "@odata.deltaLink" in graphdata:
    #             deltaLink = graphdata["@odata.deltaLink"]
    #             # yield records
    #             breakpoint()
    #         #     return odata_records, deltaLink
    #         # else:
    #         #     return odata_records, None

    async def post(
        self,
        resource: str,
        data: dict,
        endpoint: str = "v1.0",
    ):
        """Handles retry logic and token refresh for microsoft graph post request

        Post data to the graph using the odata protocol

        https://learn.microsoft.com/en-us/graph/errors

        """
        request_retry = 0
        while request_retry < 5:
            # make the request
            url = f"/{endpoint}/{resource}"
            async with self.session.post(url=url, data=data) as resp:
                if resp.status == 204:
                    resp_json = {}
                    return resp_json
                # if we get 200, return the data
                if resp.status >= 200 and resp.status < 300:
                    resp_json = await resp.json()
                    if not resp.ok:
                        log.error(f'graph error: {resp_json["error"]["message"]}')
                        log.debug(resp_json)
                    return resp_json
                # if we get 401, refresh the token and retry
                elif resp.status == 401:
                    # refresh the token
                    await self.update_token_header()
                    request_retry += 1

                # if we get 400, raise an error, something is wrong get user code to handle it
                elif resp.status == 400:
                    resp_json = await resp.json()
                    raise GraphError(resp_json["error"]["message"])

                # if we get 429, wait and retry
                elif resp.status == 429:
                    delay = float(resp.headers["retry-after"])
                    log.info(
                        f"Request to {url} failed with 429, retrying in {delay} seconds"
                    )
                    await asyncio.sleep(delay)
                    request_retry += 1
                # if we get 503, wait and retry
                elif resp.status == 503:
                    delay = float(resp.headers["retry-after"])
                    log.info(
                        f"Request to {url} failed with 503, retrying in {delay} seconds"
                    )
                    await asyncio.sleep(delay)
                    request_retry += 1
                # if we get anything else, log it and return empty
                else:
                    resp_json = await resp.json()
                    log.debug(resp_json)
                    log.error(
                        f"Request to {url} failed with {resp.status} {resp.reason}"
                    )
                    return {}

    async def delete(
        self,
        resource: str,
        endpoint: str = "v1.0",
    ):
        """Handles retry logic and token refresh for microsoft graph post request

        Delete data from the graph using the odata protocol

        https://learn.microsoft.com/en-us/graph/errors

        """
        request_retry = 0
        while request_retry < 5:
            # make the request
            url = f"/{endpoint}/{resource}"
            async with self.session.delete(url=url) as resp:
                if resp.status == 204:
                    resp_json = {}
                    return resp_json
                # if we get 200, return the data
                if resp.status >= 200 and resp.status < 300:
                    resp_json = await resp.json()
                    if not resp.ok:
                        log.error(f'graph error: {resp_json["error"]["message"]}')
                        log.debug(resp_json)
                    return resp_json
                # if we get 401, refresh the token and retry
                elif resp.status == 401:
                    # refresh the token
                    await self.update_token_header()
                    request_retry += 1

                # if we get 400, raise an error, something is wrong get user code to handle it
                elif resp.status == 400:
                    resp_json = await resp.json()
                    raise GraphError(resp_json["error"]["message"])

                # if we get 429, wait and retry
                elif resp.status == 429:
                    delay = float(resp.headers["retry-after"])
                    log.info(
                        f"Request to {url} failed with 429, retrying in {delay} seconds"
                    )
                    await asyncio.sleep(delay)
                    request_retry += 1
                # if we get 503, wait and retry
                elif resp.status == 503:
                    delay = float(resp.headers["retry-after"])
                    log.info(
                        f"Request to {url} failed with 503, retrying in {delay} seconds"
                    )
                    await asyncio.sleep(delay)
                    request_retry += 1
                # if we get anything else, log it and return empty
                else:
                    resp_json = await resp.json()
                    log.debug(resp_json)
                    log.error(
                        f"Request to {url} failed with {resp.status} {resp.reason}"
                    )
                    return {}



__doc__ = """

https://docs.djangoproject.com/en/4.0/topics/logging/
"""
import os


LOGGING_DEV = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "utc_color": {
            # NOTE: you need to include the pythonfile in the classname
            # format is set dynamically in the formatter class to include color
            "()": "odh_core.logger.UTCColorFormatter",
            # uvicorn uses need to have datefmt explicitly set
            "datefmt": "%Y-%m-%dT%H:%M:%S",
        },
        "utc": {
            "()": "odh_core.logger.UTCFormatter",
            "format": "[%(asctime)s.%(msecs)03dZ] %(levelname)s : %(name)s : %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S",
        },
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "utc_color"},
        "file": {
            # no color in log file
            "formatter": "utc",
            "filename": "odh.log",
            "class": "logging.FileHandler",
            # only log WARNING and above to file
            "level": "WARNING",
        },
    },
    "filters": {
        "hide_gql_schema": {
            "()": "odh_core.logger.HideMessageStartswith",
            "text": '<<< {"data":{"__schema":',
        },
        "hide_gql_IntrospectionQuery": {
            "()": "odh_core.logger.HideMessageStartswith",
            "text": '>>> {"query": "query IntrospectionQuery',
        },
    },
    # root sets default values and subsequent loggers can override the values
    # "root": {"handlers": ["console"], "level": "DEBUG"},
    "root": {"handlers": ["console", "file"], "level": "DEBUG"},
    "loggers": {
        "gql.transport.aiohttp": {
            "level": "DEBUG",
            "filters": ["hide_gql_schema", "hide_gql_IntrospectionQuery"],
            # "handlers": ["file"],
        },
        "odh": {
            "level": "DEBUG",
            # "handlers": ["console"],
        },
        # hide filewatcher debug messages
        "watchfiles.main": {
            "level": "WARNING",
        },
    },
}

LOGGING_PROD = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "utc_color": {
            # NOTE: you need to include the pythonfile in the classname
            # format is set dynamically in the formatter class to include color
            "()": "odh_core.logger.UTCColorFormatter",
            # uvicorn uses need to have datefmt explicitly set
            "datefmt": "%Y-%m-%dT%H:%M:%S",
        },
        "utc": {
            "()": "odh_core.logger.UTCFormatter",
            "format": "[%(asctime)s.%(msecs)03dZ] %(levelname)s : %(name)s : %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S",
        },
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "utc_color"},
        "file": {
            # no color in log file
            "formatter": "utc",
            "filename": "odh.log",
            "class": "logging.FileHandler",
            # only log WARNING and above to file
            "level": "WARNING",
        },
    },
    "filters": {
        "hide_gql_schema": {
            "()": "odh_core.logger.HideMessageStartswith",
            "text": '<<< {"data":{"__schema":',
        },
        "hide_gql_IntrospectionQuery": {
            "()": "odh_core.logger.HideMessageStartswith",
            "text": '>>> {"query": "query IntrospectionQuery',
        },
    },
    # root sets default values and subsequent loggers can override the values
    # "root": {"handlers": ["console"], "level": "DEBUG"},
    "root": {"handlers": ["console"], "level": "INFO"},
    "loggers": {
        "gql.transport.aiohttp": {
            "level": "WARNING",
            "filters": ["hide_gql_schema", "hide_gql_IntrospectionQuery"],
            # "handlers": ["file"],
        },
        "odh": {
            "level": "INFO",
            # "handlers": ["console"],
        },
        # hide filewatcher debug messages
        "watchfiles.main": {
            "level": "WARNING",
        },
    },
}

# auto set logging settings based on REMOTE_CONTAINERS env var
LOGGING_AUTO = LOGGING_PROD
if os.getenv("REMOTE_CONTAINERS", default="False").lower() in ("true", "1", "t"):
    # since vscode 1.50 remote containers will automatically
    # set env to REMOTE_CONTAINERS=true
    # see: https://github.com/microsoft/vscode-remote-release/issues/3517
    LOGGING_AUTO = LOGGING_DEV
    print("REMOTE_CONTAINERS=True, setting logging to LOGGING_DEV")

if os.getenv("LOGG_SETTING") == "PROD":
    LOGGING_AUTO = LOGGING_PROD
    print("LOGG_SETTING=PROD, setting logging to LOGGING_PROD")
elif os.getenv("LOGG_SETTING") == "DEV":
    LOGGING_AUTO = LOGGING_DEV
    print("LOGG_SETTING=DEV, setting logging to LOGGING_DEV")



from gql import Client as gql_client
from gql.transport.aiohttp import AIOHTTPTransport
from gql import gql
from graphql import DocumentNode
import logging

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


# # Tracelog for aiohttp
# # from: https://stackoverflow.com/a/64954315
# async def on_request_end(session, context, params):
#     trace_log = logging.getLogger("aiohttp.client")
#     trace_log.debug(
#         f"Request: {params.method} {params.url} "
#         f"Response: status:{params.response.status} {await params.response.text()}"
#     )

# trace_config = aiohttp.TraceConfig()
# trace_config.on_request_end.append(on_request_end)

# session = (
#     aiohttp.ClientSession(
#         base_url="https://graph.microsoft.com/",
#         headers=settings.jsonheaders,
#         trust_env=True,
#         # log all requests made by the session
#         trace_configs=[trace_config],
#     ),
# )


class GraphQLClient:
    """GraphQL client with session management

    .. todo::
        - Add support for trace logging

    """

    def __init__(
        self, url: str, headers: dict = None, ssl: bool = True, timeout: int = 10
    ):
        """Initialize GraphQL client

        Args:
            url (str): GraphQL server URL
            headers (dict, optional): HTTP headers to send with each request
            ssl (bool, optional): Verify SSL certificate
            timeout (int, optional): Timeout in seconds

        """
        self._client = gql_client(
            transport=AIOHTTPTransport(
                url=url,
                headers=headers,
                ssl=ssl,
                timeout=timeout,
            ),
            fetch_schema_from_transport=True,
        )
        self._session = None

    async def connect_async(self):
        self._session = await self._client.connect_async(reconnecting=True)
        log.info(f"Connected to GraphQL server: {self._client.transport.url}")

    async def close_async(self):
        await self._client.close_async()
        log.info(f"Connection closed to GraphQL server: {self._client.transport.url}")

    async def execute_async(self, query: DocumentNode, variable_values=None):
        """Execute a GraphQL query"""
        if not isinstance(query, DocumentNode):
            raise TypeError("query must be a gql object")

        if variable_values is None:
            result = await self._session.execute(query)
        else:
            result = await self._session.execute(query, variable_values)
        return result

    async def lock_acquire_async(self):
        """
        Create a lock in the Hasura DB, lock will release after TTL expires
        or when release is called
        """
        query = gql(
            """
            mutation InsertLock($id: String = "", $ttl_seconds: Int = 30) {
                insert_lock_table(objects: {id: $id, ttl_seconds: $ttl_seconds},
                on_conflict: {constraint: lock_table_pkey}) {
                    affected_rows
                }
            }
            """
        )
        variables = {"id": self.lock_name, "ttl_seconds": self.ttl_seconds}
        lock = await self._client.execute_async(
            query, variable_values=variables
        )
        if lock["insert_lock_table"]["affected_rows"] == 1:
            return True
        else:
            return False

    async def lock_refresh_async(self):
        """
        Create a lock in the Hasura DB, lock will release after TTL expires
        or when release is called
        """
        query = gql(
            """
            mutation RefreshLock($id: String = "", $created_at: timestamptz = now) {
                update_lock_table_by_pk(pk_columns: {id: $id}, _set: {created_at: $created_at}) {
                    id
                }
            }
            """
        )
        variables = {"id": self.lock_name}
        lock = await self._client.execute_async(
            query, variable_values=variables
        )
        if lock["update_lock_table_by_pk"]["id"] == self.lock_name:
            return True
        else:
            return False

    async def lock_release_async(self):
        """
        Release the lock in the Hasura DB
        """
        query = gql(
            """
            mutation DeleteLock($id: String = "") {
                delete_lock_table(where: {id: {_eq: $id}}) {
                    affected_rows
                }
            }
            """
        )
        variables = {"id": self.lock_name}
        lock = await self._client.execute_async(
            query, variable_values=variables
        )
        if lock["delete_lock_table"]["affected_rows"] == 1:
            return True
        else:
            return False



import phonenumbers


def parse_nr(input_phone_number, local_country="SE", fail_as_none=False) -> str | None:
    """Check for valid phone number and will always return it in pared E164 format

    Args:
        input_phone_number (str): The phonenumber
            It ignores punctuation and white-space, as
            well as any text before the number (e.g. a leading "Tel: ") and trims
            the non-number bits.  It will accept a number in any format (E164,
            national, international etc), assuming it can be interpreted with the
            defaultRegion supplied. It also attempts to convert any alpha characters
            into digits if it thinks this is a vanity number of the type "1800
            GOT MILK".

        local_country (str, optional) ISO 3166-1 alpha-2, All Caps, Default=SE (Sweden)
            This what country to try to parse local number by, This should be the country
            you are receiving most of the users from.
            see: https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements

        fail_as_none (bool, optional): Will catch error internally and return None insted

    Raises:
        ValueError if invalid phonenumber

    Return:
        Valid phonenumber in E164 format as a string

    Examples:
        >>> parse_nr("+46701234567")
        '+46701234567'
        >>> parse_nr("0701234567")
        '+46701234567'
        >>> parse_nr("1800 GOT MILK", local_country="US")
        '+18004686455'

    """
    try:
        if not isinstance(input_phone_number, str):
            raise ValueError
        nr = phonenumbers.parse(input_phone_number, region=local_country)
        if phonenumbers.is_valid_number(nr):
            return str(
                phonenumbers.format_number(nr, phonenumbers.PhoneNumberFormat.E164)
            )
        else:
            raise ValueError
    except (ValueError, phonenumbers.phonenumberutil.NumberParseException) as e:
        if fail_as_none:
            return None
        else:
            raise ValueError(f"{input_phone_number!r} is not vaild") from e
